from __future__ import annotations

import asyncio
from dataclasses import dataclass
from importlib import import_module
from typing import Any, Protocol


WalletFile = dict[str, Any]


@dataclass(slots=True)
class DilithiaAccount:
    address: str
    public_key: str
    secret_key: str
    account_index: int
    wallet_file: WalletFile | None = None


@dataclass(slots=True)
class DilithiaSignature:
    algorithm: str
    signature: str


@dataclass(slots=True)
class DilithiaKeypair:
    secret_key: str
    public_key: str
    address: str


class DilithiaCryptoAdapter(Protocol):
    def generate_mnemonic(self) -> str: ...

    def validate_mnemonic(self, mnemonic: str) -> None: ...

    def recover_hd_wallet(self, mnemonic: str) -> DilithiaAccount: ...

    def recover_hd_wallet_account(self, mnemonic: str, account_index: int) -> DilithiaAccount: ...

    def create_hd_wallet_file_from_mnemonic(self, mnemonic: str, password: str) -> DilithiaAccount: ...

    def create_hd_wallet_account_from_mnemonic(
        self, mnemonic: str, password: str, account_index: int
    ) -> DilithiaAccount: ...

    def recover_wallet_file(self, wallet_file: WalletFile, mnemonic: str, password: str) -> DilithiaAccount: ...

    def address_from_public_key(self, public_key_hex: str) -> str: ...

    def validate_address(self, addr: str) -> str: ...

    def address_from_pk_checksummed(self, public_key_hex: str) -> str: ...

    def address_with_checksum(self, raw_addr: str) -> str: ...

    def validate_public_key(self, public_key_hex: str) -> None: ...

    def validate_secret_key(self, secret_key_hex: str) -> None: ...

    def validate_signature(self, signature_hex: str) -> None: ...

    def sign_message(self, secret_key_hex: str, message: str) -> DilithiaSignature: ...

    def verify_message(self, public_key_hex: str, message: str, signature_hex: str) -> bool: ...

    def keygen(self) -> DilithiaKeypair: ...

    def keygen_from_seed(self, seed_hex: str) -> DilithiaKeypair: ...

    def seed_from_mnemonic(self, mnemonic: str) -> str: ...

    def derive_child_seed(self, parent_seed_hex: str, index: int) -> str: ...

    def constant_time_eq(self, a_hex: str, b_hex: str) -> bool: ...

    def hash_hex(self, data_hex: str) -> str: ...

    def set_hash_alg(self, alg: str) -> None: ...

    def current_hash_alg(self) -> str: ...

    def hash_len_hex(self) -> int: ...


class NativeCryptoAdapter:
    def __init__(self, module: Any) -> None:
        self._module = module

    def generate_mnemonic(self) -> str:
        return self._module.generate_mnemonic()

    def validate_mnemonic(self, mnemonic: str) -> None:
        self._module.validate_mnemonic(mnemonic)

    def recover_hd_wallet(self, mnemonic: str) -> DilithiaAccount:
        payload = self._module.recover_hd_wallet(mnemonic)
        return _normalize_account(payload)

    def recover_hd_wallet_account(self, mnemonic: str, account_index: int) -> DilithiaAccount:
        payload = self._module.recover_hd_wallet_account(mnemonic, account_index)
        return _normalize_account(payload)

    def create_hd_wallet_file_from_mnemonic(self, mnemonic: str, password: str) -> DilithiaAccount:
        payload = self._module.create_hd_wallet_file_from_mnemonic(mnemonic, password)
        return _normalize_account(payload)

    def create_hd_wallet_account_from_mnemonic(
        self, mnemonic: str, password: str, account_index: int
    ) -> DilithiaAccount:
        payload = self._module.create_hd_wallet_account_from_mnemonic(mnemonic, password, account_index)
        return _normalize_account(payload)

    def recover_wallet_file(self, wallet_file: WalletFile, mnemonic: str, password: str) -> DilithiaAccount:
        payload = self._module.recover_wallet_file(
            int(wallet_file.get("version", 1)),
            str(wallet_file.get("address", "")),
            str(wallet_file.get("public_key", wallet_file.get("publicKey", ""))),
            str(wallet_file.get("encrypted_sk", wallet_file.get("encryptedSk", ""))),
            str(wallet_file.get("nonce", "")),
            str(wallet_file.get("tag", "")),
            mnemonic,
            password,
            (
                int(wallet_file["account_index"])
                if wallet_file.get("account_index") is not None
                else None
            ),
        )
        return _normalize_account(payload)

    def address_from_public_key(self, public_key_hex: str) -> str:
        return self._module.address_from_public_key(public_key_hex)

    def validate_address(self, addr: str) -> str:
        return self._module.validate_address(addr)

    def address_from_pk_checksummed(self, public_key_hex: str) -> str:
        return self._module.address_from_pk_checksummed(public_key_hex)

    def address_with_checksum(self, raw_addr: str) -> str:
        return self._module.address_with_checksum(raw_addr)

    def validate_public_key(self, public_key_hex: str) -> None:
        self._module.validate_public_key(public_key_hex)

    def validate_secret_key(self, secret_key_hex: str) -> None:
        self._module.validate_secret_key(secret_key_hex)

    def validate_signature(self, signature_hex: str) -> None:
        self._module.validate_signature(signature_hex)

    def sign_message(self, secret_key_hex: str, message: str) -> DilithiaSignature:
        payload = self._module.sign_message(secret_key_hex, message)
        return DilithiaSignature(
            algorithm=str(payload["algorithm"]),
            signature=str(payload["signature"]),
        )

    def verify_message(self, public_key_hex: str, message: str, signature_hex: str) -> bool:
        return bool(self._module.verify_message(public_key_hex, message, signature_hex))

    def keygen(self) -> DilithiaKeypair:
        payload = self._module.keygen()
        return DilithiaKeypair(
            secret_key=str(payload["secret_key"]),
            public_key=str(payload["public_key"]),
            address=str(payload["address"]),
        )

    def keygen_from_seed(self, seed_hex: str) -> DilithiaKeypair:
        payload = self._module.keygen_from_seed(seed_hex)
        return DilithiaKeypair(
            secret_key=str(payload["secret_key"]),
            public_key=str(payload["public_key"]),
            address=str(payload["address"]),
        )

    def seed_from_mnemonic(self, mnemonic: str) -> str:
        return self._module.seed_from_mnemonic(mnemonic)

    def derive_child_seed(self, parent_seed_hex: str, index: int) -> str:
        return self._module.derive_child_seed(parent_seed_hex, index)

    def constant_time_eq(self, a_hex: str, b_hex: str) -> bool:
        return bool(self._module.constant_time_eq(a_hex, b_hex))

    def hash_hex(self, data_hex: str) -> str:
        return self._module.hash_hex(data_hex)

    def set_hash_alg(self, alg: str) -> None:
        self._module.set_hash_alg(alg)

    def current_hash_alg(self) -> str:
        return self._module.current_hash_alg()

    def hash_len_hex(self) -> int:
        return int(self._module.hash_len_hex())


def _normalize_account(payload: Any) -> DilithiaAccount:
    return DilithiaAccount(
        address=str(payload["address"]),
        public_key=str(payload["public_key"]),
        secret_key=str(payload["secret_key"]),
        account_index=int(payload.get("account_index", 0)),
        wallet_file=payload.get("wallet_file"),
    )


def load_native_crypto_adapter() -> DilithiaCryptoAdapter | None:
    try:
        module = import_module("dilithia_sdk_native")
    except ImportError:
        return None
    return NativeCryptoAdapter(module)


# ---------------------------------------------------------------------------
# Async variants
# ---------------------------------------------------------------------------


class AsyncDilithiaCryptoAdapter(Protocol):
    async def generate_mnemonic(self) -> str: ...

    async def validate_mnemonic(self, mnemonic: str) -> None: ...

    async def recover_hd_wallet(self, mnemonic: str) -> DilithiaAccount: ...

    async def recover_hd_wallet_account(self, mnemonic: str, account_index: int) -> DilithiaAccount: ...

    async def create_hd_wallet_file_from_mnemonic(self, mnemonic: str, password: str) -> DilithiaAccount: ...

    async def create_hd_wallet_account_from_mnemonic(
        self, mnemonic: str, password: str, account_index: int
    ) -> DilithiaAccount: ...

    async def recover_wallet_file(self, wallet_file: WalletFile, mnemonic: str, password: str) -> DilithiaAccount: ...

    async def address_from_public_key(self, public_key_hex: str) -> str: ...

    async def validate_address(self, addr: str) -> str: ...

    async def address_from_pk_checksummed(self, public_key_hex: str) -> str: ...

    async def address_with_checksum(self, raw_addr: str) -> str: ...

    async def validate_public_key(self, public_key_hex: str) -> None: ...

    async def validate_secret_key(self, secret_key_hex: str) -> None: ...

    async def validate_signature(self, signature_hex: str) -> None: ...

    async def sign_message(self, secret_key_hex: str, message: str) -> DilithiaSignature: ...

    async def verify_message(self, public_key_hex: str, message: str, signature_hex: str) -> bool: ...

    async def keygen(self) -> DilithiaKeypair: ...

    async def keygen_from_seed(self, seed_hex: str) -> DilithiaKeypair: ...

    async def seed_from_mnemonic(self, mnemonic: str) -> str: ...

    async def derive_child_seed(self, parent_seed_hex: str, index: int) -> str: ...

    async def constant_time_eq(self, a_hex: str, b_hex: str) -> bool: ...

    async def hash_hex(self, data_hex: str) -> str: ...

    async def set_hash_alg(self, alg: str) -> None: ...

    async def current_hash_alg(self) -> str: ...

    async def hash_len_hex(self) -> int: ...


class AsyncNativeCryptoAdapter:
    """Wraps a synchronous :class:`NativeCryptoAdapter` and delegates every
    call to a thread executor via :func:`asyncio.to_thread` so that
    CPU-intensive native crypto work never blocks the event loop.
    """

    def __init__(self, sync_adapter: NativeCryptoAdapter) -> None:
        self._sync = sync_adapter

    async def generate_mnemonic(self) -> str:
        return await asyncio.to_thread(self._sync.generate_mnemonic)

    async def validate_mnemonic(self, mnemonic: str) -> None:
        await asyncio.to_thread(self._sync.validate_mnemonic, mnemonic)

    async def recover_hd_wallet(self, mnemonic: str) -> DilithiaAccount:
        return await asyncio.to_thread(self._sync.recover_hd_wallet, mnemonic)

    async def recover_hd_wallet_account(self, mnemonic: str, account_index: int) -> DilithiaAccount:
        return await asyncio.to_thread(self._sync.recover_hd_wallet_account, mnemonic, account_index)

    async def create_hd_wallet_file_from_mnemonic(self, mnemonic: str, password: str) -> DilithiaAccount:
        return await asyncio.to_thread(self._sync.create_hd_wallet_file_from_mnemonic, mnemonic, password)

    async def create_hd_wallet_account_from_mnemonic(
        self, mnemonic: str, password: str, account_index: int
    ) -> DilithiaAccount:
        return await asyncio.to_thread(
            self._sync.create_hd_wallet_account_from_mnemonic, mnemonic, password, account_index
        )

    async def recover_wallet_file(self, wallet_file: WalletFile, mnemonic: str, password: str) -> DilithiaAccount:
        return await asyncio.to_thread(self._sync.recover_wallet_file, wallet_file, mnemonic, password)

    async def address_from_public_key(self, public_key_hex: str) -> str:
        return await asyncio.to_thread(self._sync.address_from_public_key, public_key_hex)

    async def validate_address(self, addr: str) -> str:
        return await asyncio.to_thread(self._sync.validate_address, addr)

    async def address_from_pk_checksummed(self, public_key_hex: str) -> str:
        return await asyncio.to_thread(self._sync.address_from_pk_checksummed, public_key_hex)

    async def address_with_checksum(self, raw_addr: str) -> str:
        return await asyncio.to_thread(self._sync.address_with_checksum, raw_addr)

    async def validate_public_key(self, public_key_hex: str) -> None:
        await asyncio.to_thread(self._sync.validate_public_key, public_key_hex)

    async def validate_secret_key(self, secret_key_hex: str) -> None:
        await asyncio.to_thread(self._sync.validate_secret_key, secret_key_hex)

    async def validate_signature(self, signature_hex: str) -> None:
        await asyncio.to_thread(self._sync.validate_signature, signature_hex)

    async def sign_message(self, secret_key_hex: str, message: str) -> DilithiaSignature:
        return await asyncio.to_thread(self._sync.sign_message, secret_key_hex, message)

    async def verify_message(self, public_key_hex: str, message: str, signature_hex: str) -> bool:
        return await asyncio.to_thread(self._sync.verify_message, public_key_hex, message, signature_hex)

    async def keygen(self) -> DilithiaKeypair:
        return await asyncio.to_thread(self._sync.keygen)

    async def keygen_from_seed(self, seed_hex: str) -> DilithiaKeypair:
        return await asyncio.to_thread(self._sync.keygen_from_seed, seed_hex)

    async def seed_from_mnemonic(self, mnemonic: str) -> str:
        return await asyncio.to_thread(self._sync.seed_from_mnemonic, mnemonic)

    async def derive_child_seed(self, parent_seed_hex: str, index: int) -> str:
        return await asyncio.to_thread(self._sync.derive_child_seed, parent_seed_hex, index)

    async def constant_time_eq(self, a_hex: str, b_hex: str) -> bool:
        return await asyncio.to_thread(self._sync.constant_time_eq, a_hex, b_hex)

    async def hash_hex(self, data_hex: str) -> str:
        return await asyncio.to_thread(self._sync.hash_hex, data_hex)

    async def set_hash_alg(self, alg: str) -> None:
        await asyncio.to_thread(self._sync.set_hash_alg, alg)

    async def current_hash_alg(self) -> str:
        return await asyncio.to_thread(self._sync.current_hash_alg)

    async def hash_len_hex(self) -> int:
        return await asyncio.to_thread(self._sync.hash_len_hex)


def load_async_native_crypto_adapter() -> AsyncDilithiaCryptoAdapter | None:
    try:
        module = import_module("dilithia_sdk_native")
    except ImportError:
        return None
    return AsyncNativeCryptoAdapter(NativeCryptoAdapter(module))



from __future__ import annotations

import json
import time
import asyncio
import urllib.error
import urllib.parse
import urllib.request
from typing import Any


class _DilithiaClientBase:
    def __init__(
        self,
        rpc_url: str,
        timeout: float = 10.0,
        *,
        chain_base_url: str | None = None,
        indexer_url: str | None = None,
        oracle_url: str | None = None,
        ws_url: str | None = None,
        jwt: str | None = None,
        headers: dict[str, str] | None = None,
    ) -> None:
        self.rpc_url = rpc_url.rstrip("/")
        self.base_url = (chain_base_url.rstrip("/") if chain_base_url else self.rpc_url.removesuffix("/rpc"))
        self.timeout = timeout
        self.indexer_url = indexer_url.rstrip("/") if indexer_url else None
        self.oracle_url = oracle_url.rstrip("/") if oracle_url else None
        self.ws_url = ws_url.rstrip("/") if ws_url else self._derive_ws_url()
        self.jwt = jwt
        self.headers = headers or {}

    def _derive_ws_url(self) -> str | None:
        if self.base_url.startswith("https://"):
            return "wss://" + self.base_url.removeprefix("https://")
        if self.base_url.startswith("http://"):
            return "ws://" + self.base_url.removeprefix("http://")
        return None

    def get_address_summary(self, address: str) -> dict[str, Any]:
        return self.json_rpc("qsc_addressSummary", {"address": address})

    def get_head(self) -> dict[str, Any]:
        return self.json_rpc("qsc_head", {})

    def get_chain(self) -> dict[str, Any]:
        return self.json_rpc("qsc_chain", {})

    def get_state_root(self) -> dict[str, Any]:
        return self.json_rpc("qsc_stateRoot", {})

    def get_tps(self) -> dict[str, Any]:
        return self.json_rpc("qsc_tps", {})

    def get_gas_estimate(self) -> dict[str, Any]:
        return self.json_rpc("qsc_gasEstimate", {})

    def get_base_fee(self) -> dict[str, Any]:
        return self.json_rpc("qsc_baseFee", {})

    def get_block(self, height: int) -> dict[str, Any]:
        return self.json_rpc("qsc_getBlock", {"height": height})

    def get_blocks(self, from_height: int, to_height: int) -> dict[str, Any]:
        return self.json_rpc("qsc_getBlocks", {"from": from_height, "to": to_height})

    def get_tx_block(self, tx_hash: str) -> dict[str, Any]:
        return self.json_rpc("qsc_getTxBlock", {"tx_hash": tx_hash})

    def get_internal_txs(self, tx_hash: str) -> dict[str, Any]:
        return self.json_rpc("qsc_internalTxs", {"tx_hash": tx_hash})

    def get_address_txs(self, address: str) -> dict[str, Any]:
        return self.json_rpc("qsc_getAddressTxs", {"address": address})

    def search_hash(self, hash_value: str) -> dict[str, Any]:
        return self.json_rpc("qsc_search", {"hash": hash_value})

    def build_json_rpc_request(self, method: str, params: dict[str, Any] | None = None, request_id: int = 1) -> dict[str, Any]:
        return {
            "jsonrpc": "2.0",
            "id": request_id,
            "method": method,
            "params": params or {},
        }

    def build_json_rpc_batch(self, calls: list[tuple[str, dict[str, Any]]], start_id: int = 1) -> list[dict[str, Any]]:
        return [
            self.build_json_rpc_request(method, params, request_id=index)
            for index, (method, params) in enumerate(calls, start=start_id)
        ]

    def build_network_overview_batch(self) -> list[dict[str, Any]]:
        return self.build_json_rpc_batch(
            [
                ("qsc_chain", {}),
                ("qsc_head", {}),
                ("qsc_stateRoot", {}),
                ("qsc_tps", {}),
                ("qsc_gasEstimate", {}),
            ]
        )

    def build_transaction_details_batch(self, tx_hash: str) -> list[dict[str, Any]]:
        return self.build_json_rpc_batch(
            [
                ("qsc_getReceipt", {"tx_hash": tx_hash}),
                ("qsc_getTxBlock", {"tx_hash": tx_hash}),
                ("qsc_internalTxs", {"tx_hash": tx_hash}),
            ]
        )

    def build_address_details_batch(self, address: str) -> list[dict[str, Any]]:
        return self.build_json_rpc_batch(
            [
                ("qsc_addressSummary", {"address": address}),
                ("qsc_getAddressTxs", {"address": address}),
            ]
        )

    def _parse_json_rpc_batch_response(self, body: list[dict[str, Any]], payload: list[dict[str, Any]]) -> list[dict[str, Any]]:
        items_by_id = {item["id"]: item for item in body}
        results: list[dict[str, Any]] = []
        for index, request in enumerate(payload, start=1):
            item = items_by_id[index]
            if "error" in item:
                message = item["error"].get("message", "unknown rpc error")
                raise RuntimeError(f"{request.get('method', 'rpc')} failed: {message}")
            results.append(item.get("result", {}))
        return results

    def build_ws_request(self, method: str, params: dict[str, Any] | None = None, request_id: int = 1) -> dict[str, Any]:
        return self.build_json_rpc_request(method, params, request_id)

    def build_auth_headers(self, extra: dict[str, str] | None = None) -> dict[str, str]:
        merged = {}
        if self.jwt:
            merged["Authorization"] = f"Bearer {self.jwt}"
        merged.update(self.headers)
        if extra:
            merged.update(extra)
        return merged

    def get_ws_connection_info(self) -> dict[str, Any]:
        return {"url": self.ws_url, "headers": self.build_auth_headers()}

    def build_absolute_url(self, root: str, path: str) -> str:
        return f"{root}{path if path.startswith('/') else '/' + path}"

    def build_name_service_url(self, action: str, value: str) -> str:
        return f"{self.base_url}/names/{action}/{urllib.parse.quote(value, safe='')}"

    def build_contract_query_url(self, contract: str, method: str, args: dict[str, Any] | None = None) -> str:
        encoded_args = urllib.parse.quote(json.dumps(args or {}, separators=(",", ":")), safe="")
        return (
            f"{self.base_url}/query?contract={urllib.parse.quote(contract, safe='')}"
            f"&method={urllib.parse.quote(method, safe='')}&args={encoded_args}"
        )

    def resolve_name(self, name: str) -> dict[str, Any]:
        return self._get_absolute_json(self.build_name_service_url("resolve", name))

    def lookup_name(self, name: str) -> dict[str, Any]:
        return self._get_absolute_json(self.build_name_service_url("lookup", name))

    def is_name_available(self, name: str) -> dict[str, Any]:
        return self._get_absolute_json(self.build_name_service_url("available", name))

    def get_names_by_owner(self, address: str) -> dict[str, Any]:
        return self._get_absolute_json(self.build_name_service_url("by-owner", address))

    def reverse_resolve_name(self, address: str) -> dict[str, Any]:
        return self._get_absolute_json(self.build_name_service_url("reverse", address))

    def query_contract(self, contract: str, method: str, args: dict[str, Any] | None = None) -> dict[str, Any]:
        return self._get_absolute_json(self.build_contract_query_url(contract, method, args))

    def build_forwarder_call(
        self,
        forwarder_contract: str,
        args: dict[str, Any],
        *,
        paymaster: str | None = None,
    ) -> dict[str, Any]:
        call: dict[str, Any] = {
            "contract": forwarder_contract,
            "method": "forward",
            "args": args,
        }
        return self.with_paymaster(call, paymaster) if paymaster else call

    def build_contract_call(
        self,
        contract: str,
        method: str,
        args: dict[str, Any] | None = None,
        *,
        paymaster: str | None = None,
    ) -> dict[str, Any]:
        call: dict[str, Any] = {"contract": contract, "method": method, "args": args or {}}
        return self.with_paymaster(call, paymaster) if paymaster else call

    def with_paymaster(self, call: dict[str, Any], paymaster: str) -> dict[str, Any]:
        return {**call, "paymaster": paymaster}

    def _parse_json_rpc_response(self, body: dict[str, Any], method: str) -> dict[str, Any]:
        if "error" in body:
            message = body["error"].get("message", "unknown rpc error")
            raise RuntimeError(f"{method} failed: {message}")
        return body.get("result", {})

    def _read_sync_json(self, request: urllib.request.Request) -> dict[str, Any]:
        try:
            with urllib.request.urlopen(request, timeout=self.timeout) as response:
                payload = response.read().decode("utf-8")
                return json.loads(payload)
        except urllib.error.HTTPError as exc:
            raise RuntimeError(f"HTTP {exc.code}") from exc
        except urllib.error.URLError as exc:
            raise RuntimeError(str(exc.reason)) from exc


class DilithiaClient(_DilithiaClientBase):
    def get_balance(self, address: str) -> dict[str, Any]:
        return self._get_json(f"/balance/{urllib.parse.quote(address, safe='')}")

    def get_nonce(self, address: str) -> dict[str, Any]:
        return self._get_json(f"/nonce/{urllib.parse.quote(address, safe='')}")

    def get_receipt(self, tx_hash: str) -> dict[str, Any]:
        return self._get_json(f"/receipt/{urllib.parse.quote(tx_hash, safe='')}")

    def json_rpc(self, method: str, params: dict[str, Any] | None = None, request_id: int = 1) -> dict[str, Any]:
        return self._post_json("", self.build_json_rpc_request(method, params, request_id))

    def raw_rpc(self, method: str, params: dict[str, Any] | None = None, request_id: int = 1) -> dict[str, Any]:
        return self.json_rpc(method, params, request_id)

    def json_rpc_batch(self, calls: list[tuple[str, dict[str, Any]]], start_id: int = 1) -> list[dict[str, Any]]:
        payload = self.build_json_rpc_batch(calls, start_id=start_id)
        body = self._post_json("", payload)
        return self._parse_json_rpc_batch_response(body, payload)

    def raw_get(self, path: str, *, use_chain_base: bool = False) -> dict[str, Any]:
        root = self.base_url if use_chain_base else self.rpc_url
        return self._get_absolute_json(self.build_absolute_url(root, path))

    def raw_post(self, path: str, body: Any, *, use_chain_base: bool = False) -> Any:
        root = self.base_url if use_chain_base else self.rpc_url
        return self._post_absolute_json(self.build_absolute_url(root, path), body)

    def simulate(self, call: dict[str, Any]) -> dict[str, Any]:
        return self._post_json("/simulate", call)

    def send_call(self, call: dict[str, Any]) -> dict[str, Any]:
        return self._post_json("/call", call)

    def send_signed_call(self, call: dict[str, Any], signer: Any) -> dict[str, Any]:
        payload_json = json.dumps(call, separators=(",", ":"), sort_keys=False)
        signature = signer.sign_canonical_payload(payload_json)
        if not isinstance(signature, dict):
            raise TypeError("Signer must return a dictionary payload.")
        return self.send_call({**call, **signature})

    def send_sponsored_call(self, call: dict[str, Any], paymaster: str, signer: Any) -> dict[str, Any]:
        return self.send_signed_call(self.with_paymaster(call, paymaster), signer)

    def call_contract(
        self,
        contract: str,
        method: str,
        args: dict[str, Any] | None = None,
        *,
        paymaster: str | None = None,
    ) -> dict[str, Any]:
        return self.send_call(self.build_contract_call(contract, method, args, paymaster=paymaster))

    def wait_for_receipt(self, tx_hash: str, max_attempts: int = 12, delay_seconds: float = 1.0) -> dict[str, Any]:
        for _ in range(max_attempts):
            try:
                return self.get_receipt(tx_hash)
            except RuntimeError as exc:
                if "HTTP 404" not in str(exc):
                    raise
            time.sleep(delay_seconds)
        raise RuntimeError("Receipt not available yet.")

    def _get_json(self, pathname: str) -> dict[str, Any]:
        request = urllib.request.Request(
            f"{self.rpc_url}{pathname}",
            method="GET",
            headers=self.build_auth_headers({"accept": "application/json"}),
        )
        return self._read_sync_json(request)

    def _get_absolute_json(self, url: str) -> dict[str, Any]:
        request = urllib.request.Request(
            url,
            method="GET",
            headers=self.build_auth_headers({"accept": "application/json"}),
        )
        return self._read_sync_json(request)

    def _post_json(self, pathname: str, body: Any) -> Any:
        return self._post_absolute_json(f"{self.rpc_url}{pathname}", body)

    def _post_absolute_json(self, url: str, body: Any) -> Any:
        request = urllib.request.Request(
            url,
            method="POST",
            headers=self.build_auth_headers({
                "accept": "application/json",
                "content-type": "application/json",
            }),
            data=json.dumps(body).encode("utf-8"),
        )
        return self._read_sync_json(request)


class AsyncDilithiaClient(_DilithiaClientBase):
    def __init__(
        self,
        rpc_url: str,
        timeout: float = 10.0,
        *,
        chain_base_url: str | None = None,
        indexer_url: str | None = None,
        oracle_url: str | None = None,
        ws_url: str | None = None,
        jwt: str | None = None,
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(
            rpc_url,
            timeout,
            chain_base_url=chain_base_url,
            indexer_url=indexer_url,
            oracle_url=oracle_url,
            ws_url=ws_url,
            jwt=jwt,
            headers=headers,
        )

    async def aclose(self) -> None:
        return None

    async def get_balance(self, address: str) -> dict[str, Any]:
        return await self._get_json(f"/balance/{urllib.parse.quote(address, safe='')}")

    async def get_nonce(self, address: str) -> dict[str, Any]:
        return await self._get_json(f"/nonce/{urllib.parse.quote(address, safe='')}")

    async def get_receipt(self, tx_hash: str) -> dict[str, Any]:
        return await self._get_json(f"/receipt/{urllib.parse.quote(tx_hash, safe='')}")

    async def get_address_summary(self, address: str) -> dict[str, Any]:
        return await self.json_rpc("qsc_addressSummary", {"address": address})

    async def get_head(self) -> dict[str, Any]:
        return await self.json_rpc("qsc_head", {})

    async def get_chain(self) -> dict[str, Any]:
        return await self.json_rpc("qsc_chain", {})

    async def get_state_root(self) -> dict[str, Any]:
        return await self.json_rpc("qsc_stateRoot", {})

    async def get_tps(self) -> dict[str, Any]:
        return await self.json_rpc("qsc_tps", {})

    async def get_gas_estimate(self) -> dict[str, Any]:
        return await self.json_rpc("qsc_gasEstimate", {})

    async def get_base_fee(self) -> dict[str, Any]:
        return await self.json_rpc("qsc_baseFee", {})

    async def get_block(self, height: int) -> dict[str, Any]:
        return await self.json_rpc("qsc_getBlock", {"height": height})

    async def get_blocks(self, from_height: int, to_height: int) -> dict[str, Any]:
        return await self.json_rpc("qsc_getBlocks", {"from": from_height, "to": to_height})

    async def get_tx_block(self, tx_hash: str) -> dict[str, Any]:
        return await self.json_rpc("qsc_getTxBlock", {"tx_hash": tx_hash})

    async def get_internal_txs(self, tx_hash: str) -> dict[str, Any]:
        return await self.json_rpc("qsc_internalTxs", {"tx_hash": tx_hash})

    async def get_address_txs(self, address: str) -> dict[str, Any]:
        return await self.json_rpc("qsc_getAddressTxs", {"address": address})

    async def search_hash(self, hash_value: str) -> dict[str, Any]:
        return await self.json_rpc("qsc_search", {"hash": hash_value})

    async def json_rpc(self, method: str, params: dict[str, Any] | None = None, request_id: int = 1) -> dict[str, Any]:
        return await self._post_json("", self.build_json_rpc_request(method, params, request_id), method=method)

    async def raw_rpc(self, method: str, params: dict[str, Any] | None = None, request_id: int = 1) -> dict[str, Any]:
        return await self.json_rpc(method, params, request_id)

    async def json_rpc_batch(self, calls: list[tuple[str, dict[str, Any]]], start_id: int = 1) -> list[dict[str, Any]]:
        payload = self.build_json_rpc_batch(calls, start_id=start_id)
        body = await self._post_json("", payload, method="batch")
        return self._parse_json_rpc_batch_response(body, payload)

    async def raw_get(self, path: str, *, use_chain_base: bool = False) -> dict[str, Any]:
        root = self.base_url if use_chain_base else self.rpc_url
        return await self._get_absolute_json(self.build_absolute_url(root, path))

    async def raw_post(self, path: str, body: Any, *, use_chain_base: bool = False) -> Any:
        root = self.base_url if use_chain_base else self.rpc_url
        return await self._post_absolute_json(self.build_absolute_url(root, path), body)

    async def resolve_name(self, name: str) -> dict[str, Any]:
        return await self._get_absolute_json(self.build_name_service_url("resolve", name))

    async def lookup_name(self, name: str) -> dict[str, Any]:
        return await self._get_absolute_json(self.build_name_service_url("lookup", name))

    async def is_name_available(self, name: str) -> dict[str, Any]:
        return await self._get_absolute_json(self.build_name_service_url("available", name))

    async def get_names_by_owner(self, address: str) -> dict[str, Any]:
        return await self._get_absolute_json(self.build_name_service_url("by-owner", address))

    async def reverse_resolve_name(self, address: str) -> dict[str, Any]:
        return await self._get_absolute_json(self.build_name_service_url("reverse", address))

    async def query_contract(self, contract: str, method: str, args: dict[str, Any] | None = None) -> dict[str, Any]:
        return await self._get_absolute_json(self.build_contract_query_url(contract, method, args))

    async def simulate(self, call: dict[str, Any]) -> dict[str, Any]:
        return await self._post_json("/simulate", call, method="simulate")

    async def send_call(self, call: dict[str, Any]) -> dict[str, Any]:
        return await self._post_json("/call", call, method="call")

    async def send_signed_call(self, call: dict[str, Any], signer: Any) -> dict[str, Any]:
        payload_json = json.dumps(call, separators=(",", ":"), sort_keys=False)
        signature = signer.sign_canonical_payload(payload_json)
        if not isinstance(signature, dict):
            raise TypeError("Signer must return a dictionary payload.")
        return await self.send_call({**call, **signature})

    async def send_sponsored_call(self, call: dict[str, Any], paymaster: str, signer: Any) -> dict[str, Any]:
        return await self.send_signed_call(self.with_paymaster(call, paymaster), signer)

    async def call_contract(
        self,
        contract: str,
        method: str,
        args: dict[str, Any] | None = None,
        *,
        paymaster: str | None = None,
    ) -> dict[str, Any]:
        return await self.send_call(self.build_contract_call(contract, method, args, paymaster=paymaster))

    async def wait_for_receipt(self, tx_hash: str, max_attempts: int = 12, delay_seconds: float = 1.0) -> dict[str, Any]:
        for _ in range(max_attempts):
            try:
                return await self.get_receipt(tx_hash)
            except RuntimeError as exc:
                if "HTTP 404" not in str(exc):
                    raise
            await self._sleep(delay_seconds)
        raise RuntimeError("Receipt not available yet.")

    async def _sleep(self, delay_seconds: float) -> None:
        await asyncio.sleep(delay_seconds)

    async def _get_json(self, pathname: str) -> dict[str, Any]:
        return await self._get_absolute_json(f"{self.rpc_url}{pathname}")

    async def _get_absolute_json(self, url: str) -> dict[str, Any]:
        request = urllib.request.Request(
            url,
            method="GET",
            headers=self.build_auth_headers({"accept": "application/json"}),
        )
        return await asyncio.to_thread(self._read_sync_json, request)

    async def _post_json(self, pathname: str, body: Any, *, method: str) -> Any:
        return await self._post_absolute_json(f"{self.rpc_url}{pathname}", body, method=method)

    async def _post_absolute_json(self, url: str, body: Any, *, method: str = "request") -> Any:
        request = urllib.request.Request(
            url,
            method="POST",
            headers=self.build_auth_headers({
                "accept": "application/json",
                "content-type": "application/json",
            }),
            data=json.dumps(body).encode("utf-8"),
        )
        parsed = await asyncio.to_thread(self._read_sync_json, request)
        if url == self.rpc_url and isinstance(body, dict):
            return self._parse_json_rpc_response(parsed, method)
        return parsed


class DilithiaGasSponsorConnector:
    def __init__(self, client: DilithiaClient | AsyncDilithiaClient, sponsor_contract: str, paymaster: str | None = None) -> None:
        self.client = client
        self.sponsor_contract = sponsor_contract
        self.paymaster = paymaster

    def build_accept_query(self, user: str, contract: str, method: str) -> dict[str, Any]:
        return {
            "contract": self.sponsor_contract,
            "method": "accept",
            "args": {"user": user, "contract": contract, "method": method},
        }

    def build_max_gas_per_user_query(self) -> dict[str, Any]:
        return {"contract": self.sponsor_contract, "method": "max_gas_per_user", "args": {}}

    def build_remaining_quota_query(self, user: str) -> dict[str, Any]:
        return {"contract": self.sponsor_contract, "method": "remaining_quota", "args": {"user": user}}

    def build_sponsor_token_query(self) -> dict[str, Any]:
        return {"contract": self.sponsor_contract, "method": "sponsor_token", "args": {}}

    def build_fund_call(self, amount: int) -> dict[str, Any]:
        return {"contract": self.sponsor_contract, "method": "fund", "args": {"amount": amount}}

    def apply_paymaster(self, call: dict[str, Any]) -> dict[str, Any]:
        return self.client.with_paymaster(call, self.paymaster) if self.paymaster else call

    def send_sponsored_call(self, call: dict[str, Any], signer: Any) -> Any:
        return self.client.send_signed_call(self.apply_paymaster(call), signer)


class DilithiaMessagingConnector:
    def __init__(self, client: DilithiaClient | AsyncDilithiaClient, messaging_contract: str, paymaster: str | None = None) -> None:
        self.client = client
        self.messaging_contract = messaging_contract
        self.paymaster = paymaster

    def build_send_message_call(self, dest_chain: str, payload: dict[str, Any] | str) -> dict[str, Any]:
        return self.apply_paymaster(
            {
                "contract": self.messaging_contract,
                "method": "send_message",
                "args": {"dest_chain": dest_chain, "payload": payload},
            }
        )

    def build_receive_message_call(
        self,
        source_chain: str,
        source_contract: str,
        payload: dict[str, Any] | str,
    ) -> dict[str, Any]:
        return self.apply_paymaster(
            {
                "contract": self.messaging_contract,
                "method": "receive_message",
                "args": {
                    "source_chain": source_chain,
                    "source_contract": source_contract,
                    "payload": payload,
                },
            }
        )

    def query_outbox(self) -> Any:
        return self.client.query_contract(self.messaging_contract, "outbox", {})

    def query_inbox(self) -> Any:
        return self.client.query_contract(self.messaging_contract, "inbox", {})

    def send_message(self, dest_chain: str, payload: dict[str, Any] | str, signer: Any) -> Any:
        return self.client.send_signed_call(self.build_send_message_call(dest_chain, payload), signer)

    def apply_paymaster(self, call: dict[str, Any]) -> dict[str, Any]:
        return self.client.with_paymaster(call, self.paymaster) if self.paymaster else call



from .client import AsyncDilithiaClient, DilithiaClient, DilithiaGasSponsorConnector, DilithiaMessagingConnector
from .crypto import DilithiaAccount, DilithiaCryptoAdapter, DilithiaSignature, load_native_crypto_adapter

__version__ = "0.2.0"
RPC_LINE_VERSION = "0.2.0"
MIN_PYTHON = (3, 11)

__all__ = [
    "DilithiaAccount",
    "AsyncDilithiaClient",
    "DilithiaClient",
    "DilithiaGasSponsorConnector",
    "DilithiaMessagingConnector",
    "DilithiaCryptoAdapter",
    "DilithiaSignature",
    "load_native_crypto_adapter",
]



import json
import sys
import unittest
from unittest.mock import patch

from dilithia_sdk import (
    AsyncDilithiaClient,
    DilithiaClient,
    DilithiaGasSponsorConnector,
    DilithiaMessagingConnector,
    MIN_PYTHON,
    RPC_LINE_VERSION,
    __version__,
    load_native_crypto_adapter,
)
from dilithia_sdk.crypto import NativeCryptoAdapter


class ClientTests(unittest.TestCase):
    def test_normalizes_rpc_url(self) -> None:
        client = DilithiaClient("http://rpc.example/")
        self.assertEqual(client.rpc_url, "http://rpc.example")
        self.assertEqual(client.base_url, "http://rpc.example")
        self.assertEqual(client.ws_url, "ws://rpc.example")

    def test_accepts_configurable_urls(self) -> None:
        client = DilithiaClient(
            "http://rpc.example/rpc",
            chain_base_url="http://chain.example/",
            indexer_url="http://indexer.example/",
            oracle_url="http://oracle.example/",
        )
        self.assertEqual(client.base_url, "http://chain.example")
        self.assertEqual(client.indexer_url, "http://indexer.example")
        self.assertEqual(client.oracle_url, "http://oracle.example")

    def test_builds_generic_rpc_request(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")
        self.assertEqual(
            client.build_json_rpc_request("qsc_head", {"full": True}),
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": "qsc_head",
                "params": {"full": True},
            },
        )
        self.assertEqual(
            client.build_json_rpc_batch([("qsc_head", {}), ("qsc_tps", {"window": 10})]),
            [
                {"jsonrpc": "2.0", "id": 1, "method": "qsc_head", "params": {}},
                {"jsonrpc": "2.0", "id": 2, "method": "qsc_tps", "params": {"window": 10}},
            ],
        )
        self.assertEqual(
            client.build_network_overview_batch(),
            [
                {"jsonrpc": "2.0", "id": 1, "method": "qsc_chain", "params": {}},
                {"jsonrpc": "2.0", "id": 2, "method": "qsc_head", "params": {}},
                {"jsonrpc": "2.0", "id": 3, "method": "qsc_stateRoot", "params": {}},
                {"jsonrpc": "2.0", "id": 4, "method": "qsc_tps", "params": {}},
                {"jsonrpc": "2.0", "id": 5, "method": "qsc_gasEstimate", "params": {}},
            ],
        )
        self.assertEqual(
            client.build_transaction_details_batch("deadbeef"),
            [
                {"jsonrpc": "2.0", "id": 1, "method": "qsc_getReceipt", "params": {"tx_hash": "deadbeef"}},
                {"jsonrpc": "2.0", "id": 2, "method": "qsc_getTxBlock", "params": {"tx_hash": "deadbeef"}},
                {"jsonrpc": "2.0", "id": 3, "method": "qsc_internalTxs", "params": {"tx_hash": "deadbeef"}},
            ],
        )
        self.assertEqual(
            client.build_address_details_batch("alice"),
            [
                {"jsonrpc": "2.0", "id": 1, "method": "qsc_addressSummary", "params": {"address": "alice"}},
                {"jsonrpc": "2.0", "id": 2, "method": "qsc_getAddressTxs", "params": {"address": "alice"}},
            ],
        )

    def test_explorer_rpc_helpers_delegate_to_json_rpc(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")
        calls: list[tuple[str, dict]] = []

        def fake_json_rpc(method: str, params: dict | None = None, request_id: int = 1) -> dict:
            calls.append((method, params or {}))
            return {"ok": True}

        client.json_rpc = fake_json_rpc  # type: ignore[method-assign]
        client.get_head()
        client.get_chain()
        client.get_state_root()
        client.get_tps()
        client.get_block(42)
        client.get_blocks(10, 12)
        client.get_tx_block("abc")
        client.get_internal_txs("abc")
        client.get_address_txs("alice")
        client.search_hash("deadbeef")

        self.assertEqual(
            calls,
            [
                ("qsc_head", {}),
                ("qsc_chain", {}),
                ("qsc_stateRoot", {}),
                ("qsc_tps", {}),
                ("qsc_getBlock", {"height": 42}),
                ("qsc_getBlocks", {"from": 10, "to": 12}),
                ("qsc_getTxBlock", {"tx_hash": "abc"}),
                ("qsc_internalTxs", {"tx_hash": "abc"}),
                ("qsc_getAddressTxs", {"address": "alice"}),
                ("qsc_search", {"hash": "deadbeef"}),
            ],
        )

    def test_sync_json_rpc_batch_parses_results(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")

        def fake_post_json(_pathname: str, body):
            self.assertIsInstance(body, list)
            return [
                {"jsonrpc": "2.0", "id": 1, "result": {"chain_id": "dili-devnet"}},
                {"jsonrpc": "2.0", "id": 2, "result": {"height": 42}},
            ]

        client._post_json = fake_post_json  # type: ignore[method-assign]
        result = client.json_rpc_batch([("qsc_chain", {}), ("qsc_head", {})])
        self.assertEqual(result, [{"chain_id": "dili-devnet"}, {"height": 42}])

    def test_jwt_auth_headers_are_exposed_for_http_and_ws(self) -> None:
        client = DilithiaClient(
            "https://rpc.example/rpc",
            jwt="secret-token",
            headers={"x-network": "devnet"},
        )
        self.assertEqual(
            client.build_auth_headers({"accept": "application/json"}),
            {
                "Authorization": "Bearer secret-token",
                "x-network": "devnet",
                "accept": "application/json",
            },
        )
        self.assertEqual(
            client.get_ws_connection_info(),
            {
                "url": "wss://rpc.example",
                "headers": {
                    "Authorization": "Bearer secret-token",
                    "x-network": "devnet",
                },
            },
        )

    def test_sdk_version_stays_aligned_with_rpc_line(self) -> None:
        self.assertEqual(__version__, "0.2.0")
        self.assertEqual(RPC_LINE_VERSION, "0.2.0")

    def test_python_runtime_satisfies_minimum_supported_version(self) -> None:
        self.assertGreaterEqual(sys.version_info[:2], MIN_PYTHON)

    def test_native_crypto_adapter_loader_returns_none_when_bridge_is_unavailable(self) -> None:
        self.assertIsNone(load_native_crypto_adapter())

    def test_native_crypto_adapter_maps_bridge_module_correctly(self) -> None:
        class FakeModule:
            @staticmethod
            def generate_mnemonic() -> str:
                return "word " * 24

            @staticmethod
            def validate_mnemonic(_mnemonic: str) -> None:
                return None

            @staticmethod
            def address_from_public_key(_public_key_hex: str) -> str:
                return "derived-address"

            @staticmethod
            def sign_message(_secret_key_hex: str, _message: str) -> dict:
                return {"algorithm": "mldsa65", "signature": "deadbeef"}

            @staticmethod
            def verify_message(_public_key_hex: str, _message: str, _signature_hex: str) -> bool:
                return True

            @staticmethod
            def recover_wallet_file(*_args) -> dict:
                return {
                    "address": "addr-1",
                    "public_key": "pk-1",
                    "secret_key": "sk-1",
                    "account_index": 0,
                    "wallet_file": {"version": 1},
                }

        adapter = NativeCryptoAdapter(FakeModule)
        self.assertEqual(adapter.address_from_public_key("pk"), "derived-address")
        self.assertEqual(adapter.sign_message("sk", "msg").signature, "deadbeef")
        self.assertTrue(adapter.verify_message("pk", "msg", "sig"))
        recovered = adapter.recover_wallet_file({"version": 1}, "mnemonic", "password")
        self.assertEqual(recovered.address, "addr-1")
        self.assertEqual(recovered.public_key, "pk-1")
        self.assertEqual(recovered.secret_key, "sk-1")

    def test_native_crypto_adapter_loader_uses_imported_module_when_available(self) -> None:
        class FakeModule:
            pass

        with patch("dilithia_sdk.crypto.import_module", return_value=FakeModule):
            adapter = load_native_crypto_adapter()
        self.assertIsNotNone(adapter)

    def test_send_signed_call_merges_signer_payload(self) -> None:
        client = DilithiaClient("http://rpc.example")

        class FakeSigner:
            @staticmethod
            def sign_canonical_payload(payload_json: str) -> dict:
                return {"sig": f"signed:{payload_json}", "pk": "pk", "alg": "mldsa65"}
        

        sent = {}

        def fake_send_call(call: dict) -> dict:
            sent.update(call)
            return {"tx_hash": "0xabc"}

        client.send_call = fake_send_call  # type: ignore[method-assign]
        result = client.send_signed_call({"from": "a", "contract": "token"}, FakeSigner())
        self.assertEqual(result, {"tx_hash": "0xabc"})
        self.assertEqual(sent["pk"], "pk")
        self.assertEqual(sent["alg"], "mldsa65")
        self.assertTrue(sent["sig"].startswith("signed:"))

    def test_paymaster_and_forwarder_helpers(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")
        sponsored = client.with_paymaster({"contract": "wasm:amm", "method": "swap"}, "gas_sponsor")
        self.assertEqual(sponsored["paymaster"], "gas_sponsor")

        forwarder = client.build_forwarder_call(
            "wasm:forwarder",
            {"user": "alice", "nonce": 1},
            paymaster="gas_sponsor",
        )
        self.assertEqual(forwarder["contract"], "wasm:forwarder")
        self.assertEqual(forwarder["method"], "forward")
        self.assertEqual(forwarder["paymaster"], "gas_sponsor")

    def test_name_service_and_contract_query_use_chain_base_url(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")
        called_urls: list[str] = []

        def fake_read_json(_request):
            called_urls.append(_request.full_url)
            return {"ok": True}

        client._read_sync_json = fake_read_json  # type: ignore[method-assign]
        client.resolve_name("alice.dili")
        client.query_contract("wasm:amm", "get_reserves", {})
        self.assertEqual(called_urls[0], "http://rpc.example/names/resolve/alice.dili")
        self.assertEqual(
            called_urls[1],
            "http://rpc.example/query?contract=wasm%3Aamm&method=get_reserves&args=%7B%7D",
        )

    def test_gas_sponsor_connector_builds_expected_queries(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")
        sponsor = DilithiaGasSponsorConnector(client, "wasm:gas_sponsor", paymaster="gas_sponsor")
        self.assertEqual(
            sponsor.build_remaining_quota_query("alice"),
            {"contract": "wasm:gas_sponsor", "method": "remaining_quota", "args": {"user": "alice"}},
        )
        applied = sponsor.apply_paymaster({"contract": "wasm:amm", "method": "swap", "args": {}})
        self.assertEqual(applied["paymaster"], "gas_sponsor")

    def test_messaging_connector_builds_in_and_out_calls(self) -> None:
        client = DilithiaClient("http://rpc.example/rpc")
        messaging = DilithiaMessagingConnector(client, "wasm:messaging", paymaster="gas_sponsor")
        outbound = messaging.build_send_message_call("ethereum", {"amount": 1})
        self.assertEqual(outbound["method"], "send_message")
        self.assertEqual(outbound["paymaster"], "gas_sponsor")
        inbound = messaging.build_receive_message_call("ethereum", "bridge", {"tx": "0xabc"})
        self.assertEqual(inbound["method"], "receive_message")
        self.assertEqual(inbound["args"]["source_chain"], "ethereum")


class AsyncClientTests(unittest.IsolatedAsyncioTestCase):
    async def test_async_client_exposes_jwt_for_ws(self) -> None:
        client = AsyncDilithiaClient("https://rpc.example/rpc", jwt="token", headers={"x-network": "devnet"})
        self.assertEqual(
            client.get_ws_connection_info(),
            {
                "url": "wss://rpc.example",
                "headers": {
                    "Authorization": "Bearer token",
                    "x-network": "devnet",
                },
            },
        )
        await client.aclose()

    async def test_async_json_rpc_and_batch_paths(self) -> None:
        client = AsyncDilithiaClient("http://rpc.example/rpc", jwt="token")
        calls: list[tuple[str, dict | list[dict], str]] = []

        async def fake_post_json(pathname: str, body, *, method: str):
            calls.append((pathname, body, method))
            if isinstance(body, list):
                return [
                    {"jsonrpc": "2.0", "id": 1, "result": {"chain_id": "dili-devnet"}},
                    {"jsonrpc": "2.0", "id": 2, "result": {"height": 42}},
                ]
            return {"height": 42}

        async def fake_post_absolute_json(url: str, body, *, method: str = "request"):
            return await fake_post_json(url, body, method=method)

        client._post_json = fake_post_json  # type: ignore[method-assign]
        client._post_absolute_json = fake_post_absolute_json  # type: ignore[method-assign]
        head = await client.get_head()
        batch = await client.raw_post("", client.build_json_rpc_batch([("qsc_chain", {}), ("qsc_head", {})]))

        self.assertEqual(head, {"height": 42})
        self.assertEqual(batch[0]["result"], {"chain_id": "dili-devnet"})
        self.assertEqual(calls[0][2], "qsc_head")

    async def test_async_json_rpc_batch_parses_results(self) -> None:
        client = AsyncDilithiaClient("http://rpc.example/rpc")

        async def fake_post_json(_pathname: str, body, *, method: str):
            self.assertEqual(method, "batch")
            self.assertIsInstance(body, list)
            return [
                {"jsonrpc": "2.0", "id": 1, "result": {"chain_id": "dili-devnet"}},
                {"jsonrpc": "2.0", "id": 2, "result": {"height": 42}},
            ]

        client._post_json = fake_post_json  # type: ignore[method-assign]
        result = await client.json_rpc_batch([("qsc_chain", {}), ("qsc_head", {})])
        self.assertEqual(result, [{"chain_id": "dili-devnet"}, {"height": 42}])


if __name__ == "__main__":
    unittest.main()
